/* gfClient - A client for the genomic finding program that produces a .psl file. */
#include "common.h"
#include "linefile.h"
#include "aliType.h"
#include "fa.h"
#include "genoFind.h"
#include "psl.h"
#include "cheapcgi.h"

void Make_RC_Seq(char **orig_seq, char **rc_seq, int size);

/* Variables that can be overridden by command line. */
int dots = 0;
int minScore = 20;
double minIdentity = 0;
char *outputFormat = "psl";

void outputOptions(struct gfSavePslxData *outForm, FILE *f)
/* Set up main output data structure. */
{
ZeroVar(outForm);
outForm->f =f;
outForm->minGood = round(10*minIdentity);
outForm->saveSeq = sameWord(outputFormat, "pslx");
}

void gfClient(char **servers, int num_servers, char *nibDir, char *inName, 
	char *outName, char *tTypeName, char *qTypeName)
/* gfClient - A client for the genomic finding program that produces a .psl file. */
{
struct lineFile *lf = lineFileOpen(inName, TRUE);
static bioSeq seq;
struct ssBundle *bundleList;
FILE *out = mustOpen(outName, "w");
enum gfType qType = gfTypeFromName(qTypeName);
enum gfType tType = gfTypeFromName(tTypeName);
int dotMod = 0;
char *hostName, *portName;
int i, conn;

if (!cgiVarExists("nohead"))
    pslWriteHead(out);
while (faSomeSpeedReadNext(lf, &seq.dna, &seq.size, &seq.name, qType != gftProt))
  {
    DNA *orig_seq = seq.dna;
    DNA *rc_seq = NULL;
    for (i = 0; i < num_servers; ++i) {
      hostName = servers[i * 2];
      portName = servers[i * 2 + 1];
      conn = gfConnect(hostName, portName);
      if (dots != 0)
        {
	if (++dotMod >= dots)
	    {
	    dotMod = 0;
	    fputc('.', stdout);
	    fflush(stdout);
	    }
	}
      if (qType == gftProt && (tType == gftDnaX || tType == gftRnaX))
        {
	static struct gfSavePslxData outForm;
	outputOptions(&outForm, out);
	outForm.reportTargetStrand = TRUE;
	outForm.qIsProt = TRUE;
	gfAlignTrans(conn, nibDir, &seq, minScore, gfSavePslx, &outForm);
	}
      else if ((qType == gftRnaX || qType == gftDnaX) && (tType == gftDnaX || tType == gftRnaX))
        {
	static struct gfSavePslxData outForm;
	outputOptions(&outForm, out);
	outForm.reportTargetStrand = TRUE;
	seq.dna = orig_seq;
	gfAlignTransTrans(conn, nibDir, &seq, FALSE, minScore, gfSavePslx, &outForm, qType == gftRnaX);
	if (qType == gftDnaX)
	    {
	    close(conn);
	    if (num_servers > 1)
	      Make_RC_Seq(&seq.dna, &rc_seq, seq.size);
	    else
	      reverseComplement(seq.dna, seq.size);
	    conn = gfConnect(hostName, portName);
	    gfAlignTransTrans(conn, nibDir, &seq, TRUE, minScore, gfSavePslx, &outForm, FALSE);
	    }
	}
      else if ((tType == gftDna || tType == gftRna) && (qType == gftDna || qType == gftRna))
	{
	static struct gfSavePslxData outForm;
	outputOptions(&outForm, out);
	seq.dna = orig_seq;
	gfAlignStrand(conn, nibDir, &seq, FALSE, minScore, gfSavePslx, &outForm);
	close(conn);
	if (num_servers > 1)
          Make_RC_Seq(&seq.dna, &rc_seq, seq.size);
	else
	  reverseComplement(seq.dna, seq.size);
	conn = gfConnect(hostName, portName);
	gfAlignStrand(conn, nibDir, &seq, TRUE,  minScore, gfSavePslx, &outForm);
	}
      else
        {
	errAbort("Comparisons between %s queries and %s databases not yet supported",
		qTypeName, tTypeName);
	}
      close(conn);
    }
    free(rc_seq);
  }
if (out != stdout)
    printf("Output is in %s\n", outName);
}

void Make_RC_Seq(char **orig_seq, char **rc_seq, int size) {
  if (*rc_seq == NULL) {
    *rc_seq = malloc(size + 1);
    if (*rc_seq == NULL)
      errAbort("Couldn't allocate memory for rc_seq");
    strcpy(*rc_seq, *orig_seq);
    reverseComplement(*rc_seq, size);
  }
  *orig_seq = *rc_seq;
}

void usage()
/* Explain usage and exit. */
{
errAbort(
  "gfClient - A client for the genomic finding program that produces a .psl file\n"
  "usage:\n"
  "   gfClient host port [host2 port2 ...] nibDir in.fa out.psl\n"
  "where\n"
  "   host is the name of the machine running the gfServer\n"
  "   port is the same as you started the gfServer with\n"
  "   nibDir is the path of the nib files relative to the current dir\n"
  "       (note these are needed by the client as well as the server)\n"
  "   in.fa a fasta format file.  May contain multiple records\n"
  "   out.psl where to put the output\n"
  "options:\n"
  "   -t=type     Database type.  Type is one of:\n"
  "                 dna - DNA sequence\n"
  "                 prot - protein sequence\n"
  "                 dnax - DNA sequence translated in six frames to protein\n"
  "               The default is dna\n"
  "   -q=type     Query type.  Type is one of:\n"
  "                 dna - DNA sequence\n"
  "                 rna - RNA sequence\n"
  "                 prot - protein sequence\n"
  "                 dnax - DNA sequence translated in six frames to protein\n"
  "                 rnax - DNA sequence translated in three frames to protein\n"
  "   -dots=N   Output a dot every N query sequences\n"
  "   -nohead   Suppresses psl five line header");
}

int main(int argc, char *argv[])
/* Process command line. */
{
cgiFromCommandLine(&argc, argv, FALSE);
if ((argc < 6) || ((argc % 2) != 0))
    usage();
minIdentity = cgiUsualDouble("minIdentity", minIdentity);
minScore = cgiOptionalInt("minScore", minScore);
dots = cgiOptionalInt("dots", 0);
outputFormat = cgiUsualString("out", outputFormat);
gfClient(argv + 1, (argc - 4) / 2, argv[argc - 3], argv[argc - 2],
  argv[argc - 1], cgiUsualString("t", "dna"), cgiUsualString("q", "dna"));
return 0;
}
